
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Analysis</h1>
#         <h3 style="text-align:center;">Getting Started: Exploratory Data Analysis (EDA)</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> EDA, matplotlib, tokenization, glob, Counters, paths
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# This notebook is adapted from one by Ted Underwood of the University of Illinois.

# # Reading text and counting it
# 
# We will eventually learn how to abbreviate some of these operations with the ```nltk``` module, but first we must learn how datasets are formed and transformed.

# ### Absolute and Relative Paths
# 
# When you're working at the command line, your location in the folder hierarchy is represented by a series of directory names separated by slashes. If you're in OS X, for instance and you type
# 
#     > pwd
# 
# you'll get a string that represents the current "working directory," like,
# 
#     > /Users/jonathanplamb/Dropbox/Digital/CRMDA/StatsCamp/software_python_summer/python-1/Python-1-2-EDA
# 
# In Windows, the slashes go the other way, so this might be
# 
#     > cd
#     C:\Users\jonathanplamb\Dropbox\Digital\CRMDA\StatsCamp\software_python_summer\python-1\Python-1-2-EDA
#    
# This is called a *path.* In this case, it's a path to a directory. If we wanted to name a specific file in the directory, we could provide the path plus a filename, such as
# 
#     /Users/jonathanplamb/Dropbox/Digital/CRMDA/StatsCamp/software_python_summer/python-1/Python-1-2-EDA/Python-1-2-EDA-texts.ipynb
#    
# which is the location of this file on my machine.
# 
# These are *absolute paths* because they start from the root of the folder hierarchy and contain all the information needed to locate the file. In writing this notebook, however, I don't know your user name, or exactly where you've put our course folder on your machine. So I'm going to try to use *relative paths* to point to files.
# 
# In this case, I know that your python notebooks are in a folder named ```Python-1-2-EDA```, which is itself in a folder named ```Python-1```. The data files for our workshop are in a folder called ```data```, which is next to the  ```Python-1``` folder. So if I wanted to point to a file in the ```data``` folder, I might say,
# 
#     ../../data/exampledata.txt
# 
# You'll recall that ```cd ..``` moves you up a level in the folder hierarchy. So the function of the two ```..```s in the relative path is to move us up two levels in the hierarchy, to the parent folder. Then ```data``` moves us back down to the data folder, and ```exampledata.txt``` is an (imaginary) data file.
# 
# The only problem with this system is that it won't work on Windows. The slashes would need to go the other way. So let's make use of a special Python tool that can get around that hassle, the *module* ```os```. A module is a collection of pieces of Python code. If you want to use one or more of those pieces, you can *import* the module. The ```os``` module contains tools that can be used to manipulate your operating system. So, for instance,
# 

# In[ ]:


import os
relativepath = os.path.join('..', '..', 'data', 'Corpus', 'MobyDick.txt')
print(relativepath)


# The result you get by running that code will depend on the operating system you're running. If you run it under Windows, you'll get slashes pointing the other direction.
# 
# Now let's use this relative path to open a file
# 
# ### Text Input
# 
# When you're reading or writing from a file, it's important to close the file when you're done. There are ways to do that manually, but generally it's simpler and more reliable to use Python's *context manager,* an indented block of code that begins with the key word ```with```. For instance,
# 
# 

# In[ ]:


import os
relativepath = os.path.join('..', '..', 'data', 'Corpus', 'MobyDick.txt')

list_of_lines = list()

with open(relativepath, mode = 'r', encoding = 'utf-8') as f:
    for line in f:
        print(line)
        user_wish = input('continue? ')
        if user_wish == 'n':
            break
        else:
            list_of_lines.append(line)
        
print(len(list_of_lines))    


# If you wanted to gather the lines all at once, you wouldn't need a loop. Instead you could just say
# 
#     list_of_lines == f.readlines()
# 
# ### Tokenizing Text.
# 
# Right now we're looking at the file as a list of lines. But in reality, to do data science on text, we usually want to count individual words — or, word-like objects (numbers, etc). We can call these objects generally *tokens,* and the process of transforming text into a list of tokens is called *tokenizing.*
# 
# In the last class we met the 'split' function, which gets us part of the way there, relying on the fact that English words are generally divided by white space.

# In[ ]:


firstline = 'The Project Gutenberg EBook of Moby Dick; or The Whale, by Herman'
tokens = firstline.split()
print(tokens)


# That's a list of tokens, but not exactly a list of words. You'll notice, for instance, that there's punctuation attached to some of the words.
# 
# But still it's good enough to give us a rough estimate of the number of words in a file.
# 
# ### Counting tokens.
# 
# Let's count the total number of tokens in *Moby Dick.* For the moment, we'll define a token simply as a sequence of characters separated by whitespace.

# In[ ]:


import os
relativepath = os.path.join('..', '..', 'data', 'Corpus', 'MobyDick.txt')

number_of_tokens = 0

with open(relativepath, mode = 'r', encoding = 'utf-8') as f:
    for line in f:
        tokens = line.split()
        number_of_tokens += len(tokens)

print('The file contained {0} tokens.'.format(number_of_tokens))


# Notice that strings have a ```format``` method that can be used when you need to provide variables to "fill in" certain "blanks."

# ### The ```glob``` module!
# 
# Suppose we wanted to know the number of words for all the files in the ```Corpus``` folder. Instead of writing a separate piece of code for each one, we would probably use a loop. So we need,
# 
# * a list of all the files in the folder, and
# * then, for each file, do the same thing we did above for *Moby Dick*
# 
# How can we do that?? We might make use of the ```glob``` module. Then, once we have a list of paths to all the files, we build a new ```for``` loop and pasting some code from above. Count the words for each file, and print the number of words next to the file path.

# In[ ]:


import os, glob
relativepath = os.path.join('..', '..', 'data', 'Corpus', '*.txt')
print(relativepath)
textfilepaths = glob.glob(relativepath)
print(textfilepaths)


# In[ ]:


for apath in textfilepaths:
    number_of_tokens = 0

    with open(apath, mode = 'r', encoding = 'utf-8') as f:
        for line in f:
            tokens = line.split()
            number_of_tokens += len(tokens)
    print(apath, number_of_tokens)


# ## Word Frequencies
# 
# Counting the *total* number of words in a text doesn't tell us very much (especially when the words we have seen don't actually appear in *Moby Dick*). Often, when we're working with text, we'll want to know the frequencies of *specific* words.
# 
# When we start counting words, we're going to need to produce a data structure that's a little more complex than a list. We need a *mapping* between words and numbers. So, for instance, here are some frequencies in this paragraph:
# 
#     a : 4
#     here : 1
#     going : 1
#     so : 1
#     for : 1
#     to : 2
#     words : 2
# 
# When we need to represent a *mapping* in Python, we usually create a *dictionary.* A dictionary maps a key to a value, as we saw earlier.

# In[ ]:


colors = dict()
colors['apple'] = 'red'
colors['banana'] = 'yellow'
print(colors)
print(colors['banana'])


# Putting a key in square brackets after the dictionary name gets you the value for that key. But notice what happens if you try a key that hasn't been entered.

# In[ ]:


colors['guava']


# Notice also that the brackets have to be square. If you use parentheses, Python will think you're trying to *call* a function.

# In[ ]:


colors('apple')


# We can count word frequencies by mapping each word to the number of times it has occurred in the text. We iterate through the text, and each time we see a word, we add one to its value in the wordcount dictionary. The only problem is, we can't add one if it's not yet in the dictionary! We would get a key error, as seen above. So when you use a dictionary to count things, you have to check whether the key is already in the dictionary. An if-else statement is handy here. Follow the logic below.

# In[ ]:


paragraph = "When we start counting words, we're going to need to produce a data structure that's a little more complex than a list. We need a *mapping* between words and numbers. So, for instance, here are some frequencies in this paragraph:"
wordcounts = dict()
for w in paragraph.split():
    w = w.lower()
    w = w.strip('\n,.* "\'')
    if w not in wordcounts:
        wordcounts[w] = 1
    else:
        wordcounts[w] = wordcounts[w] + 1

# You cannot iterate through a dictionary quite as simply as if it were
# a list. Instead of individual elements, you want (key, value) pairs.
# To get them you can iterate through the dictionary's ```items``` method.
# Note that a method has to be followed by () parentheses, even when it
# takes no argument.

for word, count in wordcounts.items():
    print(word + " : " + str(count))


# Now, try deleting the line
# 
#     w = w.lower().strip(',.* ')
#     
# and see how the results change.

# To save ourselves the labor of checking whether a word is in the dictionary yet, we can use a ```Counter,``` a convenient data structure in the ```collections``` module. Instead of returning a KeyError when you ask for a key that doesn't exist yet, a Counter returns 0. So you can skip the if-else and just start by adding one to the default value of 0.
# 
# Another nice feature of ```Counters``` is that they have a ```most_common()``` method that gives you a sorted list of the keys with the highest values. In the code below, we use that feature to identify the five most frequent words.

# In[ ]:


from collections import Counter

paragraph = "When we start counting words, we're going to need to produce a data structure that's a little more complex than a list. We need a *mapping* between words and numbers. So, for instance, here are some frequencies in this paragraph:"
wordcounts = Counter()
for w in paragraph.split():
    w = w.lower().strip(',.* ')
    wordcounts[w] = wordcounts[w] + 1

topfive = wordcounts.most_common(5)
print(topfive)


# Notice that each element of the list ```topfive``` is a pair of items enclosed in parentheses. This is actually an example of a new data structure, a **tuple.** A tuple is basically an immutable list. You can't append to it, but there are some nice tricks you can perform, especially something called *tuple packing*, which allows you to turn a series of items separated by a comma into a single object, and then unpack it again into multiple objects. 

# In[ ]:


a_tuple = ('need', 2)
word, frequency = a_tuple
print(word)


# ### Exercise! Word frequencies.
# 
# Adapt your code from Exercise 1 so that, instead of just counting the total number of words in *Moby Dick,* you identify the most common 100 words. Include a line like
# 
#     w = w.lower().strip(',.*"-—() ')
# 
# so that we don't count words separately when they're only distinguished by capitalization or punctuation.
# 
# Put the list of ```most_common()``` words in a list called ```tophundred```. Print the first ten items from ```tophundred```.
# 

# In[ ]:


from collections import Counter
relativepath = os.path.join('..', '..', 'data', 'Corpus', 'MobyDick.txt')

wordcounts = Counter()

with open(relativepath, mode = 'r', encoding = 'utf-8') as f:
    for line in f:
        for w in line.split():
            w = w.lower().strip(',.* ()-_–[]')
            wordcounts[w] = wordcounts[w] + 1

tophundred = wordcounts.most_common(100)

print(tophundred[0:10])


# ## III. A bit of exploratory data visualization.
# 
# Now we can start to explore the data, to get a sense of its shape. One simple thing to do is to plot the raw word counts for the first hundred words. To do this we need to use a module called ```matplotlib```.
# 
# We have to say ```%matplotlib inline``` as a kind of magic word that permits graphs to be displayed in a Jupyter notebook. Then we import a specific subset of matplotlib called "pyplot." We rename it ```plt```, for a shorter convenience name. Then we pass ```plt```'s plot method a list of numbers, which it treats as the y axis values. (If x values aren't provided, it assumes by default that we want a sequence of integers, which in this case is exactly what we do want.) Finally, we have to say ```plt.show()``` to make the image actually appear.

# This crude visualization is already revealing something pretty fundamental about the way word frequencies behave. Try the same exercise with *Pride and Prejudice.* (You'll have to rerun Exercise 3, using a different file name, and then re-run the visualization code above.)
# 
# You have just discovered [Zipf's Law.](https://en.wikipedia.org/wiki/Zipf's_law)

# In[ ]:


get_ipython().magic('matplotlib inline')
import matplotlib.pyplot as plt

justcounts = [x[1] for x in tophundred]
plt.plot(justcounts)
plt.ylabel('raw count of occurrences')
plt.xlabel('ordinal position of word in ranked list')
plt.show()


#     ro = red circles
#     bo = blue circles
#     k+ = black crosses
#     g-- = green dashes
#     ms = magenta squares   

# In[ ]:


plt.xlim((-5, 105))
# that option defines the limits of the x axis; here it's
# set manually to give us extra space on either side

plt.plot(justcounts, 'ro')

plt.ylabel('raw count of occurrences')
plt.xlabel('ordinal position of word in ranked list')

plt.grid(True)
# turning on grid lines
plt.show()


# Play around a bit with the code above to change the color and shape of the points, the axis labels, and the limits of the y or x axes.

# We've learned a lot in this Notebook! In the next one, we'll use more features and continue to add to our working knowledge of Python.
