
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Getting Started: Working with Language (pt. 1)</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> text, txt, NLTK, natural language processing
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # Getting Started with Language

# In this notebook, we will focus on working with language-as-data. As we did previously, we'll be building from the ground up, starting with the very basics to get a grasp of the concepts and then working our way up to more advanced forms of analysis. Some of today's work has been heavily adapted from tutorials by Geoffrey Rockwell and Stefan Sinclair.
# 
# ***
# 
# In lieu of a lengthy conceptual and theoretical orientation to working with language-as-data, consider the etymology of the word *data*. It comes from the Latin word meaning "to give." Data is regarded as **something given**, as a thing that simply comes to us to be analyzed and used as something given to us.
# 
# ### But this is not quite true!!
# 
# As [Johanna Drucker argues](http://www.digitalhumanities.org/dhq/vol/5/1/000091/000091.html), "data" is never something we are simply given. Rather, what we call "data" is in fact more like "capta"--something taken or captured. As you begin working with data (however seemingly neutral or innocuous), keep this principle mind. It will mean two key things:
# 1. It will mean putting a lot of thought into *what exactly you're working with when you work with data.* You will have to think hard about what you're counting, what you're excluding, and what it means to count or process certain things in certain ways.
# 2. It will refine the claims you make with computational work. A robust suspicion toward data will undermine any naive belief that data-driven claims can ever merely "tell the truth."
# In today's lessons, we will encounter several instances of the consequences of this principle. From the very first line of code, data analysis carries many assumptions and values that affect how we treat data and what kinds of claims we can make with it. That said, however, this principle does not mean we can't do intresting and illuminating work using computational analysis!
# 
# To begin, we'll learn how to grab text files from the internet and save them. Then we'll get in the weeds of language.

# In[ ]:


import urllib.request

# We create a variable for the Project Gutenberg plaintext version
# of Frederick Douglass's Narrative. Then we grab that file and assign it 
# to a string variable called narrataive_string.

narrative_URL = "http://www.gutenberg.org/files/23/23.txt"
narrative_string = urllib.request.urlopen(narrative_URL).read().decode().strip()


# In[ ]:


print(narrative_string)


# In[ ]:


print("This string has", len(narrative_string), "characters.")


# In[ ]:


# Let's print the first 25 characters of the narrative string!
narrative_string[:25]


# In[ ]:


# A reminder about working with strings using 

print("first character:", narrative_string[0])
print("last character:", narrative_string[-1])
print("first 25 characters:", narrative_string[:25])
print("last 25 characters:", narrative_string[-25:])
print("characters 95 through 208", narrative_string[95:208])


# For more, see the Python [string documentation](https://docs.python.org/3/library/stdtypes.html#string-methods).
# 
# ### How about some basic functions?

# In[ ]:


print("Occurrences of 'colonel':", narrative_string.count('colonel'))


# In[ ]:


# What do the commands below tell us about how the count() method works?

print("Occurrences of 'colonel':", narrative_string.count('colonel'))
print("Occurrences of 'colone':", narrative_string.count('colonel'))
print("Occurrences of 'Colonel':", narrative_string.count('Colonel'))


# In[ ]:


# That's easy enough to fix, though not sustainable.
# Adding a series of methods is called "function chaining."

print("Occurrences of 'COLONEL':", narrative_string.upper().count('COLONEL'))


# Now let's say we want to extract just the text of Douglass's Narrative. When we printed the string above, we saw that Project Gutenberg includes lots of boiler plate language at the beginning and end of the file. If we can point to where the text begins and ends, we can extract it.
# 
# Where does Douglass's Narrative begin and end? (This is one of those moments where the notion of DATA is vexed.)

# In[ ]:


# Once we decide where Douglass's Narrative starts and ends,
# insert the relevant strings in the quotation marks below.
# BEWARE: the .find() method will find the first matching case.

start = narrative_string.find("")
end = narrative_string.find("")
print(narrative_string[start:end].strip())


# In[ ]:


# Can you see the potential problem in the code below?

narrative_string = narrative_string[start:end].strip()
print(narrative_string[:50], "[...]", narrative_string[-50:])


# Ok, so we've extracted just the text of the Narrative and assigned it to the variable ```narrative_string```. Now how can we save this to a file?
# 
# First, remember what we learned before about navigating filepaths. We need to point from where we are (in a folder called "Python-2-1-Texts") up **two levels** and then **down one level** in to the "data" folder. In that folder, we want to create a directory (i.e., a folder) called "Douglass". We will talk through this step-by-step.

# In[ ]:


# What is my Present Working Directory?

pwd


# In[ ]:


import os
relativepath = os.path.join('..', '..', 'data', 'Douglass')
print(relativepath)

# If there's no Douglass folder, make one
if not os.path.exists(relativepath):
    os.makedirs(relativepath)


# Next, we want to open a new text file called 'narrative.txt' and write our ```narrative_string``` string variable to the file. We already know some of the steps:

# In[ ]:


# First create a filepath to point to where we want the file to go.

filepath = os.path.join(relativepath, 'narrative.txt')

# Then use the 'with' command to open the file in WRITE mode (hence 'w')

with open(filepath, 'w') as f:
    f.write(narrative_string)


# That's it! If you check in the folder, you should have a file called 'narrative.txt'. It contains the extracted text of Douglass's Narrative. If we want, we can open the file and read it into a new string:

# In[ ]:


with open(filepath, 'r') as f:
    douglass_string = f.read()

print(douglass_string)


# # NLTK for Text Processing

# NLTK, or [Natural Language Toolkit](http://www.nltk.org/), facilitates a range of tasks for text processing and representing results. It's especially useful in combination with other Python packages. We won't cover all of the features of NLTK, but we'll highlight some of the most common ones.
# 
# First, you need to download the NLTK data on your computer. Anaconda comes pre-packaged with NLTK, but we need the data collections too. **Running the cell below will open a new window in your computer prompting you to download.** When it opens (and it'll take a minute), download all packages.

# In[ ]:


import nltk

nltk.download() # download NLTK data (we should only need to run this cell once)


# Now we will work toward showing the top frequency words in our plain text. This involves three major steps:
# 
# 1. processing our plain text to find the words (also known as tokenization)
# 1. counting the frequencies of each word
# 1. displaying the frequencies information

# ## Tokenization and Cleanup
# Tokenization is the basic process of parsing a string to divide it into smaller units of the same kind. You can tokenize text into paragraphs, sentences, words or other structures, but here we're focused on recognizing words in our text. For that, let's import the ```nltk``` library and use its convenient ```word_tokenize()``` function. NLTK actually has several ways of tokenizing texts, and for that matter we could write our own code to do it. We'll have a peek at the first ten tokens.

# In[ ]:


import nltk

douglass_tokens = nltk.word_tokenize(douglass_string)
douglass_tokens[:100]


# It's clear that we need to do some cleanup here. There are lots of ways to accomplish that cleanup, all dependent on our goals and on the assumptions we are willing to carry forward into the data. For instance, we could simply iterate over the list of tokens and remove anything that does not include an alphanumeric character in the first slot:

# In[ ]:


print([word for word in douglass_tokens[:100] if word[0].isalpha()])


# In[ ]:


# Or we could try it this way. What's the difference?
# What difference does that difference make?

print([word for word in douglass_tokens[:100] if word.isalpha()])


# There seem to be problems with both methods! For some text files, this simple method would work, but because of the nature of the data, we must try a bit harder. How can we remove the punctuation we don't want (commas, etc.) but retain the puncuation we do (hyphenated words, contractions)? We'll have to create a...
