
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Working with Language</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> text, txt, NLTK, natural language processing
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # Function!
# We have already made use of lots of functions, which are more or less like commands. (There are also methods, which are slightly different in that they are attached to particular classes of objects.) But let's say we want a way to do the same task over and over again--how can we define our own function to accomplish this? It's quite easy:
# 
# ```python 
#     def [function name](*arguments)
# ```
#     
# That's it! We define a function (hence, ```def```), give it a name, and specify what information we want to pass to it. Let's grab our string of Douglass's *Narrative* and then define some functions for removing punctuation.

# In[ ]:


import os
import nltk

# First create a filepath to point to where we want the file to go.

relativepath = os.path.join('..', '..', 'data', 'Douglass')
filepath = os.path.join(relativepath, 'narrative.txt')

# Then use the 'with' command to open the file in WRITE mode (hence 'w')

with open(filepath, 'r') as f:
    narrative_string = f.read()


# In[ ]:


print(narrative_string)


# In[ ]:


# Method 1: we iterate through our list of punctuation markers,
# and use the replace() method to replace the marker with nothing.
# Note that we have to be selective about which punctuation we want to remove.
# For instance, the '--' must be removed by itself, while we have an opion
# to remove apostrophes (as in possessives or contractions).

def remove_punc(text):
    text = text.replace('--', ' ')
    #punctuation = '!@#$%^&*()_-+={}[]:;"\'|<>,.?/~`'
    punctuation = '!@#$%^&*()_-+={}[]:;"|<>,.?/~`'
    for marker in punctuation:
        text = text.replace(marker, "")
    return text


# you can play around, inputting different strings here, to see what the function will strip.
short_text = "A slaveholder's profession of Christianity is a palpable imposture."
print(remove_punc(short_text))


# In[ ]:


# Method 2: We iterate through the text itself, and if the character
# is not a punctuation mark, then we add it to the 'clean_text' string.
# Again, we have to remove the '--' character by itself.

def remove_punc2(text):
    text = text.replace('--', ' ')
    #punctuation = '!@#$%^&*()_-+={}[]:;"\'|<>,.?/~`'
    punctuation = '!@#$%^&*()_-+={}[]:;"|<>,.?/~`'
    clean_text = ""
    for character in text:
        if character not in punctuation:
            clean_text += character
    return clean_text

short_text2 = "Fortunate, most fortunate occurrence!--fortunate for the millions of his manacled brethren, yet panting for deliverance from their awful thraldom!--fortunate for the cause of negro emancipation, and of universal liberty!--fortunate for the land of his birth, which he has already done so much to save and bless!--fortunate for a large circle of friends and acquaintances, whose sympathy and affection he has strongly secured by the many sufferings he has endured, by his virtuous traits of character, by his ever-abiding remembrance of those who are in bonds, as being bound with them!--"
print(remove_punc2(short_text2))


# In[ ]:


# Method 3. we use regular expressions, which we'll over on Day 4.

import re

def remove_punc3(text):
    text = text.replace('--', ' ')
    text = re.sub(r'[^\w\s]','', text)
    return(text)

short_text3 = "Fortunate, most fortunate occurrence!--fortunate for the millions of his manacled brethren, yet panting for deliverance from their awful thraldom!--fortunate for the cause of negro emancipation, and of universal liberty!--fortunate for the land of his birth, which he has already done so much to save and bless!--fortunate for a large circle of friends and acquaintances, whose sympathy and affection he has strongly secured by the many sufferings he has endured, by his virtuous traits of character, by his ever-abiding remembrance of those who are in bonds, as being bound with them!--"
print(remove_punc3(short_text3))


# #### Discussion:
# What are the pros and cons of each method?
# 1. 
# 2. 
# 3. 
# 
# ***
# 
# So now we have a function to call, and a text variable to which we want to send it. (And note that we are still working with a **string**, not (yet) a list of tokens.

# In[ ]:


narrative_string_nopunct = remove_punc(narrative_string)
print(narrative_string_nopunct)

# Well, have we removed all the punctuation?
# What's at stake in removing the punctuation? (Look back at short_text3)


# What now? If we tokenize this string now, we'll have another problem: capitalization! We have a method just for that. Once we make the string lowercase, we're ready to tokenize.

# In[ ]:


# Lowercase the whole Narrative string.

clean_narrative_string = narrative_string_nopunct.lower()

# Tokenize!

import nltk
narrative_tokens = nltk.word_tokenize(clean_narrative_string)
print(narrative_tokens[:25])


# ### Word Frequencies
# Now that we've had a first pass at word tokenization (keeping only word tokens), let's look at counting word frequencies. Essentially we want to go through the tokens and tally the number of times each one appears. Not surprisingly, the NLTK has a very convenient method for doing just this, which we can see in this small sample (the first 50 word tokens):

# In[ ]:


narrative_tokens_sample = narrative_tokens[:50]
sample_freqs = nltk.FreqDist(narrative_tokens_sample)
sample_freqs


# This FreqDist object is a kind of dictionary, where each word is paired with its frequency (separated by a colon), and each pair is separated by a comma. This kind of dictionary also has a very convenient way of displaying results as a table:

# In[ ]:


sample_freqs.tabulate(10)


# Now let's do it again with the whole text of Douglass's *Narrative*.

# In[ ]:


narrative_freqs = nltk.FreqDist(narrative_tokens)
narrative_freqs.tabulate(10)


# One simple way of measuring the vocabulary richness of an author is to calculate the ratio of the total number of words and the number of unique words. If an author repeats words more often, it may be because he or she is drawing on a smaller vocabulary (either deliberately or not), which is a measure of style. There are several factors to consider, such as the length of the text--the longer the text, the more likely someone is to repeat words more frequently--but in the simplest terms we can calculate the lexical diversity of an author by dividng the number of word forms (types) by the total number of tokens. We already have the necessary information:
# 
# - types: number of different words (number of words: count pairs in narrative_freqs)
# - tokens: total number of word tokens (length of narrative_tokens)

# In[ ]:


print("number of types: ", len(narrative_freqs))
print("number of tokens: ", len(narrative_tokens))
print("type/token ratio: ", len(narrative_freqs)/len(narrative_tokens))


# **Question**: In what situations would this information be meaningful? What are its limitations?

# ### More cleanup?
# 
# There are several other things we might want to 'clean' out of this file. Depending on our goals, we may want to:
# - Remove numbers
# - Remove common words (called stopwords)
# - Remove textual markers (e.g., 'chapter i')

# In[ ]:


# To remove numbers, we simply iterate over the list and remove anything
# without an alphabetical first character

narrative_tokens_nonumbers = [word for word in narrative_tokens if word.isalpha()]


# In[ ]:


# To remove stopwords, we import a list of stopwords from NLTK,
# then simply iterate over the list and include only words that are NOT stopwords

from nltk.corpus import stopwords
stopwords = nltk.corpus.stopwords.words("english")
narrative_tokens_nostop = [word for word in narrative_tokens_nonumbers if word not in stopwords]
print(narrative_tokens_nostop[:50])


# We now have a list of words in Douglass's *Narrative* that does not include punctuation, stopwords, numbers, and is lowercased. This process is typically called 'cleaning.' Can you think of any reasons this metaphor may not be quite right?
# 
# Possible issues notwithstanding, we can now look at the frequency of Douglass's non-stopword vocabulary:

# In[ ]:


narrative_nostop_freqs = nltk.FreqDist(narrative_tokens_nostop)
narrative_nostop_freqs.tabulate(20)


# ### Building a simple condordance
# 
# A concordance allows us to see each occurrence of a term in its context. It has a rich history in textual scholarship, dating back to well before the advent of computers. It's a tool for studying word usage in context.
# 
# The easiest way to build a concordance is to create an NLTK Text object from a list of word tokens (in this case we'll use the unfiltered list so that we can better read the text). So, for instance, we can ask for a concordance of "mr" to try to better understand why it occurs so often in this text. Insert other words into the concordance function and see what you can find.

# In[ ]:


narrativeText = nltk.Text(narrative_tokens)
narrativeText.concordance('mr', lines=50)


# ## Some Graphing

# In[ ]:


# make sure that graphs are embedded into our notebook output
get_ipython().magic('matplotlib inline')

# plot the top frequency words in a graph
narrative_freqs.plot(25, title="Top Freqency Tokens Including Stopwords")
narrative_nostop_freqs.plot(25, title="Top Frequency Word Tokens Exclusing Stopwords")


# The first graph shows not only the rank of the words (along the bottom x axis), but is also much more effective than the table at showing the steep decline in frequency as we move away from the first words. This is actually a well-known phenomenon with natural language and is described by Zipf's law, which the Wikipedia article nicely summarizes:
# 
#     Zipf's law states that given some corpus of natural language utterances, the frequency of any word is inversely proportional to its rank in the frequency table. Thus the most frequent word will occur approximately twice as often as the second most frequent word, three times as often as the third most frequent word, etc.
# 
# As we continue to explore frequency of words, it's useful to keep in mind the distinction between frequency rank and the actual number of words (tokens) that each word form (type) is contributing.
# 
# In the second graph we made, you can see the frequency of particular terms. As we might expect in Douglass's narrative, the words 'slave' and 'master' appear with great frequency. We might want to use a dispersion plot to visualize where they appear in the text. As with the concordance, we call up the NLTK Text object (rather than any of the lists of tokens we've created).

# In[ ]:


narrativeText.dispersion_plot(['slave', 'master'])


# ### Parts of Speech
# Parts of speech are linguistic categories that identify the syntactic function of words, such as determiners (or articles), nouns, verbs, adjectives, prepositions, etc. Imagine, for instance, that we could determine that Text A has more adjectives per total number of words than Text B – might that allow us to speculate that Text A is more descriptive than Text B? Or, to increase the scale, imagine that we could compare the parts of speech of 1,000 texts (or 1 million?).
# 
# Let's first look at the mechanics of identifying – or tagging – parts of speech in our text. We'll be using the ```narrative_tokens``` file, which includes stopwords and numbers but not punctuation.

# In[ ]:


import nltk

narrative_tagged = nltk.pos_tag(narrative_tokens)
narrative_tagged[:25]


# Tagging this way is an "expensive" operation, which is a computer programming way of saying that it's an intensive process that takes time and resources (like URL fetching, we want to do it as infrequently as possible). Behind the scenes, NLTK is loading a large dataset that it's using to help improve the tagging results.
# 
# We see here a list of tuples with each tuple being composed of a token from the text and a code representing the part-of-speech. The codes are listed at the Penn Treebank Project and includes the following:
# 
#     DT: determiner (the, a…)
#     IN: preposition (to, from, …)
#     JJ: adjective (big, smart…)
#     NN: noun (car, school…)
#     NNP: proper noun (Montreal, Jill…)
#     PRP: personal pronoun (she, I…)
#     RB: adverb (freely, calmly…)
#     VBD, VBG, VBN, VBZ: verbs (dancing, said, been, is…)
#     WP: wh-pronoun (who, what…)
# 
# Looking back at our initial list, we can see there are some correct tags, but others that seem dubious. For instance, 'frederick' is incorrectly tagged as an adjective and 'antislavery' is tagged as a noun.
# 
# It's worth reiterating that the computer doesn't understand anything about the language being used, it's making a statistical guess based on the training set that it's already seen. In other words, based on a corpus of eclectic texts, the part-of-speech tagger we're using is identifying certain words as, say, determiners, because that's how they were tagged previously. Some words can be a different part-of-speech depending on context (dance as a verb or dance as a noun), and the tagging may depend on words that precede or follow. Some words it simply won't recognize and may just guess at depending on the adjacent tags (or a default tag).
# 
# For relatively normal modern English texts we can expect an accuracy of about 85-90%, which is far from perfect, but which does allow us to explore the texts in new ways. The accuracy is probably lower for Douglass's *Narrative*, which was not modern. (And please note here the language of 'normalcy' is another of those pesky problems with the question of data!)
# 
# We may want to filter for (or filter out!) a particular part of speech, which is easy enough:

# In[ ]:


narrative_nouns = [token for token, pos in narrative_tagged if "NN" in pos]
nltk.FreqDist(narrative_nouns).plot(20)


# Each part of speech might lead us to some insights, and it might be valuable to look at the overall tag frequencies:

# In[ ]:


get_ipython().magic('matplotlib inline')
tags = [pos for word, pos in narrative_tagged]
tags_freqs = nltk.FreqDist(tags)
tags_freqs.tabulate(20)
tags_freqs.plot(20)


# In[ ]:


# Do we learn anything from a dispersion plot of the tags?

short_tags = [pos[0:2] for pos in tags]
short_tagged_freqs = nltk.FreqDist(short_tags)
nltk.Text(short_tags).dispersion_plot(list(short_tagged_freqs.keys()))


# ### N-Grams

# So far we've been focused on considering the frequency and distribution of individual terms, but sometimes it can also be useful to study repeating phrases, or sequences of words. Repeating phrases are sometimes also called n-grams </a> where the "n" indicates a variable number of repeating words (or other linguistic unit like a character or ideogram). So we can have unigrams (one), bigrams (two), trigrams (three), and so on. The Google Ngram Viewer provides a way of searching for variable length phrases in its very large corpus of digitized works. N-Grams have proven instrumental in determining stylistic peculiarities and other unconscious language choices.
# 
# We will again be using the ```narrative_tokens``` variable, which contains numbers and stopwords, but not punctuation.

# In[ ]:


import nltk
narrative_bigrams = list(nltk.ngrams(narrative_tokens, 2))
narrative_bigrams[:5]


# As we can see, the nltk.ngrams() function goes through the text, and creates a new tupple of the specified length, with each item shifting forward by just one word – this is a moving window that slides along the entire text.
# 
# Since the last word of the text can't be the start of a bigram, we should have one fewer bigrams than words (even if all words but the last word are repeated in two entries, as the first word and as the second word).

# In[ ]:


print("number of words:", len(narrative_tokens))
print("number of ngrams:", len(narrative_bigrams))


# In[ ]:


narrative_bigrams_freqs = nltk.FreqDist(narrative_bigrams)

for words, count in narrative_bigrams_freqs.most_common(20):
    print(count, " ".join(list(words)))


# In[ ]:


narrative_4grams = list(nltk.ngrams(narrative_tokens, 4))
narrative_4grams_freqs = nltk.FreqDist(narrative_4grams)

for words, count in narrative_4grams_freqs.most_common(20):
    print(count, " ".join(list(words)))


# In[ ]:


get_ipython().magic('matplotlib inline')
narrative_4grams_text = nltk.Text(narrative_4grams)
narrative_4grams_text.dispersion_plot([('at', 'the', 'same', 'time')])


# What if we want to find the longest repeating phrase in the *Narrative*? This may (or may not) be useful.

# In[ ]:


ngramsFreqs = [] # keep track of last set of repeating phrases
for length in range(2, len(narrative_tokens)): # create a range from 2 to the length of the text (even if we don't get that high)
    ngrams = list(nltk.ngrams(narrative_tokens, length)) # get n-grams for specified length
    freqs = nltk.FreqDist(ngrams) # get frequencies for these n-grams
    freqs = [(ngram, count) for ngram, count in freqs.items() if count > 1] # filter out frequencies that don't repeat
    if len(freqs) > 0: # we have at least one repeating phrase
        ngramsFreqs = freqs # new set of frequencies
    else: # if we've filtered out all frequencies then break out of the loop
        break

for ngram, count in ngramsFreqs:
    print("ngram of ", len(ngram), " words occurring ", count, " times:", " ".join(list(ngram)))


# ### Segmenting Texts
# 
# We've relied heavily on dispersion_plot() as a way of looking at term distribution, but there are other approaches, ones that may be better or worse depending on the context. One possibility is to segment our text into equal parts and then to plot distribution for each part. (This is also useful for more advanced analysis, so pay attention!)
# 
# We can segment our tokens or n-grams in various ways, but one convenient way is to use a library called NumPy that is helpful for working with arrays (and graphing array-backed data). Let's see it in action with a simplified example where we're going to create a list of numbers from 0 to 9 (10 elements) and then create 5 equal bins in which to place the items.

# In[ ]:


numbers = list(range(0,9)) # create a range from 0 to 9 and convert the range to a list
numbers


# In[ ]:


import numpy as np
numberBins = np.array_split(numbers, 5) # divide our numbers into 5 equal bins
print("number of bins:", len(numberBins))
numberBins


# We can do the same thing to our bigrams of the *Narrative*, splitting it into 10 'bins'.

# In[ ]:


# First we have to convert the bigrams (which is a list of tuples)
# into joined tokens.
narrative_bigrams_tokens = [' '.join(gram) for gram in narrative_bigrams]

# Then we segment those tokens in a numpy array of ten bins
narrative_bigrams_segments = np.array_split(narrative_bigrams_tokens, 10)

#How many bigrams in each bin?
[len(segment) for segment in narrative_bigrams_segments] 


# Now, instead of counting the number of tokens, we'll count the number of occurrences of a search term--let's say "a slave".

# In[ ]:


a_slave_counts = [list(segment).count(('a slave')) for segment in narrative_bigrams_segments]
a_slave_counts


# In[ ]:


import matplotlib.pyplot as plt
line = plt.plot(a_slave_counts, label="a slave")
plt.ylim(0) # make sure to set y axis to start at zero
plt.legend(handles=line) # we're going to add a legend
plt.show() # make sure to flush out the line chart


# In[ ]:


import matplotlib.pyplot as plt
xaxis = range(0, len(a_slave_counts)) # a range from zero to the length of our counts (0-9)
bar = plt.bar(xaxis, a_slave_counts, label="a slave")
plt.legend(handles=[bar])
plt.show()


# In[ ]:


searches = ["of the", "in the"]
lines = []
for search in searches:
    line, = plt.plot([list(segment).count(search) for segment in narrative_bigrams_segments], label=search)
    lines.append(line)
plt.legend(handles=lines)
plt.show()


# # Having fun yet?
# We've covered a lot in this part of the workshop. In the next session, we'll build on the foundations we've built here and begin to explore texts at greater levels of abstraction.
