
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Working with Language: Digital Text Analysis</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> collocates, semantic analysis
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # Next steps
# We have covered some basics of working with texts in Python. In this session, we will explore some more advanced methods, keeping in mind questions of scale (what works for 1 text can be multiplied to more texts) and data-curation (data is not neutral!).

# ## Getting Collocates for Words
# 
# One very simple and common computational method is finding collocates.

# ### Checking files and getting a text
# 
# First we check what texts we have in the directory.

# Copy in the title of the text you want to process and run the next cell. Our example uses the plain text version of Hume's [A Treatise of Human Nature by David Hume](http://www.gutenberg.org/ebooks/4705).

# In[ ]:


import os

filepath = os.path.join('..', '..', 'data', 'Hume_Treatise.txt')

with open(filepath, "r") as f:
    theText = f.read()

print("This string has", "{:,}".format(len(theText)), "characters")


# ### Tokenizing
# 
# I have taken the liberty of removing the Project Gutenberg licensing information and table of contents, so we should be working with 'the text'--whatever that is.

# In[ ]:


import re # Regular expressions again! We'll get to those on Day 4.
theTokens = re.findall(r'\b\w[\w-]*\b', theText.lower())
print(theTokens[:10])


# ### Finding the collocates
# 
# This will ask you what word you want collocates for create a list off collocates. Note that you can set the number of words of context.

# In[ ]:


wrd2find = input("What word do you want collocates for?") # Ask for the word to search for
context = 5 # This sets the context of words on either side to grab

end = len(theTokens)
counter = 0
theCollocates = []
for word in theTokens:
    if word == wrd2find: # This checks to see if the word is what we want
        for i in range(context):
            if (counter - (i + 1)) >= 0: # This checks that we aren't at the beginning
                theCollocates.append(theTokens[(counter - (i + 1))]) # This adds words before
            if (counter + (i + 1)) < end: # This checks that we aren't at the end
                theCollocates.append(theTokens[(counter + (i + 1))]) # This adds words after
    counter = counter + 1
    
print(theCollocates[:10])


# In[ ]:


print(theCollocates)


# ### Doing things with the collocates
# 
# Now we can do various things with the list of collocates.

# #### Count collocates

# In[ ]:


print(len(theCollocates))


# #### Count unique words among collocates

# In[ ]:


print(set(theCollocates))


# #### Tabulate top collocates

# In[ ]:


import nltk
tokenDist = nltk.FreqDist(theCollocates)
tokenDist.tabulate(10)


# #### Plot top collocates

# In[ ]:


import matplotlib
get_ipython().magic('matplotlib inline')
tokenDist.plot(25, title="Top Frequency Collocates for " + wrd2find.capitalize())


# #### Export all collocates and counts as a CSV
# 
# This will create CSV file with the name of the target word with the counts.

# In[ ]:


import csv
nameOfResults = wrd2find.capitalize() + ".Collocates.csv"
output_filepath = os.path.join('..','..','data',nameOfResults)
table = tokenDist.most_common()

with open(output_filepath, "w") as f:
    writer = csv.writer(f)
    writer.writerows(table)
    
print("Done")


# # Sentiment Analysis in two parts
# We're going to walk through two versions of how to do sentiment analysis. This may seem redundant, but each method presents different outputs for this kind of analysis.
# 
# Sentiment analysis, in short, uses pre-determined scores to count the sentiment for a text or set of texts. Words (and sometimes other things) receive 'sentiment' scores, and texts are scored using different methods for their sentiment.
# 
# Our first example uses Shakespeare's sonnets. 
# ## The first step is to create a corpus of the sonnets.

# In[ ]:


import urllib.request
sonnetsURL = "http://www.gutenberg.org/files/1041/1041.txt"
sonnetString = urllib.request.urlopen(sonnetsURL).read().decode()


# In[ ]:


print(sonnetString)


# In[ ]:


# We need to trim the text file, using the output above.
start = sonnetString.find("  I\r\n")
end = sonnetString.find("End of Project Gutenberg")
filteredSonnetString = sonnetString[start:end].rstrip()
print(filteredSonnetString)


# In[ ]:


# Next we need to filter the sonnets and create a list of them.
import re
sonnetsList = re.split("  [A-Z]+\r\n\r\n", filteredSonnetString)
print(sonnetsList[1])


# In[ ]:


# Now we create a directory for saving our sonnets.

import os
sonnetsPath = os.path.join('..','..','data','sonnets')
if not os.path.exists(sonnetsPath):
    os.makedirs(sonnetsPath)


# In[ ]:


# Next we create a file name for each item in the list
# and write the file to a new .txt file.

for index, sonnet in enumerate(sonnetsList):
    if len(sonnet.strip()) > 0:
        filename = str(index).zfill(3)+".txt"
        pathname = os.path.join(sonnetsPath, filename)
        f = open(pathname, "w")
        f.write(sonnet.strip())
        f.close()

# If you check your 'data' folder, there should be a 'sonnets' 
# folder with 154 files in it, each conntaining a single sonnet.


# In[ ]:


# The reason we did all of that work is that nltk has a 
# useful corpus reader, which we'll use to 'read' the files
# into a corpus object.

from nltk.corpus import PlaintextCorpusReader

sonnetsCorpus = PlaintextCorpusReader(sonnetsPath, '.*txt')
len(sonnetsCorpus.fileids())


# In[ ]:


# Let's write a little function to give us a summary of this corpus.
# This function could of course be reused on any corpus.

def corpus_summary(corpus):
    print("This corpus has")
    print("  ", '{:,}'.format(len(sonnetsCorpus.fileids())), "files")
    tokens = corpus.words()
    print("  ", '{:,}'.format(len(tokens)), "tokens")
    words = [word for word in tokens if word[0].isalpha()]
    print("  ", '{:,}'.format(len(words)), "words")
    print("  ", '{:,}'.format(len(set(words))), "unique words")

corpus_summary(sonnetsCorpus)


# ## Simple Sentiment Analysis

# In[ ]:


# Now we tokenize! The nltk.corpus words() method does this for us.

corpusTokens = sonnetsCorpus.words()
print(corpusTokens[:10])


# We'll begin the process with some relatively straightforward sentiment scores, just to illustrate how this works.

# In[ ]:


# Can you tell us what kind object we're creating?
# How do you know?

short_sentiments = {
    "pos": {
        "love": 1,
        "like": .5
    },
    "neg": {
        "hate": -1,
        "dislike": -.5
    }
}


# In[ ]:


# Here we score each sonnet with our short sentiment scores.
# Let's walk through this script step by step.

import nltk
documents = {}
for fileid in sonnetsCorpus.fileids():
    text = sonnetsCorpus.raw(fileid).lower()
    tokens = nltk.word_tokenize(text)
    score = 0
    for polarity, words_dict in short_sentiments.items():
        for word, value in words_dict.items():
            score += tokens.count(word) * value
    documents[fileid] = score
    


# In[ ]:


# How did the sonnets score? Let's use a FreqDist to check.

valuesFreqs = nltk.FreqDist(documents)
valuesFreqs.tabulate(200)


# In[ ]:


mostFreqField = valuesFreqs.max()
print(sonnetsCorpus.raw(mostFreqField))


# ## Slightly more complex sentiment analysis
# Ok, we have the basic sense of how this works. NLTK contains a feature called ```sentiwordnet```, which scores English words in much greater detail. Let's see how this works. NLTK stores this information in a structure called a *synset*, which we won't spend much time discussing.
# 
# **Be warned:** it gets a little wonky here, but stick with it to the end.

# In[ ]:


from nltk.corpus import sentiwordnet as swn

# What scores does the letter 'd' have?
# Replace 'd' with another word and have a look at its scores.
for senti_synset in swn.senti_synsets('d'):
    print(senti_synset, senti_synset.synset.definition())


# In[ ]:


# We can also pass a part of speech (here 'a').
# The more specific we can be with what part of speech
# a word is, the more accurate our sentiment scores will be.

for senti_synset in swn.senti_synsets('wicked', 'a'):
    print(senti_synset, senti_synset.synset.definition())


# Here's where we get in the weeds for a minute.
# 
# The problem is that NLTK's part-of-speech tags come in Treebank form (such as "NN"), while the synsets come in Wordnet form (such as 'a'). So we have to convert one to the other before we can score a tagged text.

# In[ ]:


text = "This is a good sentence."
tokens = nltk.word_tokenize(text)
tagged = nltk.pos_tag(tokens)
tagged


# In[ ]:


# Let's write a handy function for converting the part of speech tags.
def treebank_to_wordnet_pos(treebank, skipWordNetPos=[]):
    if "NN" in treebank and "n" not in skipWordNetPos:
        return "n"
    elif "JJ" in treebank and "a" not in skipWordNetPos:
        return "a"
    elif "VB" in treebank and "v" not in skipWordNetPos:
        return "v"
    elif "RB" in treebank and "r" not in skipWordNetPos:
        return "r"


# In[ ]:


# Now let's run our sample sentence through the converter.

for word, treebank in tagged:
    # The 'v' is included here as an argument, thus the function excludes it
    wordnet_pos = treebank_to_wordnet_pos(treebank, ["v"]) 
    if wordnet_pos:
        print(word, wordnet_pos)


# Now we need a function to which we can send a tagged set of tokens and Treebank part of speech tag.

# In[ ]:


def get_sentiment_score_from_tagged(token, treebank, skipWordNetPos=[]):
    wordnet_pos = treebank_to_wordnet_pos(treebank, skipWordNetPos)
    if wordnet_pos: # only print matches
        senti_synsets = list(swn.senti_synsets(token, wordnet_pos))
        if senti_synsets:
            return senti_synsets[0].pos_score() - senti_synsets[0].neg_score()


# In[ ]:


# Now let's try it on our sample sentence.

for word, treebank in tagged:
    score = get_sentiment_score_from_tagged(word, treebank, ["v"])
    if score:
        print(word, score)
        
# Can you explain the output? Why did it return only the words it returned?


# Ok, so now we can (1) convert the part-of-speech tags (excluding the ones we don't want) and (2) get a sentiment score from a part-of-speech-tagged set of tokens. **But**, we want a way to send a set of tokens and retrieve an overall score, along with a list of the positive words and negative words. We can write yet another function.

# In[ ]:


def get_sentiment_data_from_tokens(tokens, skipWordNetPos=[]):
    tagged = nltk.pos_tag(tokens)
    positives = []
    negatives = []
    tokens_score = 0
    for word, treebank in tagged:
        score = get_sentiment_score_from_tagged(word, treebank, skipWordNetPos)
        if score:
            tokens_score += score
            if score > 0:
                positives.append(word.lower())
            else:
                negatives.append(word.lower())
    return tokens_score, set(positives), set(negatives)


# In[ ]:


get_sentiment_data_from_tokens(tokens)


# Ok, one step more. We don't just want to send a single set of tokens. What if we want to send a **corpus** of texts, like the sonnet corpus we've created? We need one final function:

# In[ ]:


def get_sentiments_data_from_corpus(corpus, skipWordNetPos=[]):
    documents = {}
    all_positives = []
    all_negatives = []
    for fileid in corpus.fileids():
        tokens = corpus.words(fileid)
        score, positives, negatives = get_sentiment_data_from_tokens(tokens, skipWordNetPos)
        documents[fileid] = score
        [all_positives.append(positive) for positive in positives]
        [all_negatives.append(negative) for negative in negatives]
    return documents, set(all_positives), set(all_negatives)


# In[ ]:


# Now we can simply call this function, sending it our sonnet corpus!
sonnetsSentimentValues, sonnetsPositives, sonnetsNegatives = get_sentiments_data_from_corpus(sonnetsCorpus)


# In[ ]:


# What do the data look like?
print(sonnetsSentimentValues)


# In[ ]:


sonnetsSentimentFreqs = nltk.FreqDist(sonnetsSentimentValues)
sonnetsSentimentFreqs.tabulate(10)
get_ipython().magic('matplotlib inline')
sonnetsSentimentFreqs.plot()


# ## Another method of sentiment analysis

# Another method, which has different strengths and weaknesses, is to use a pre-determined list of positive and negative words to score the corpus. This notebook is adapted from Neal Caron's [An introduction to text analysis with Python, Part 3](http://nealcaren.web.unc.edu/an-introduction-to-text-analysis-with-python-part-3/).
# 
# First we need to read our positive and negative dictionaries, both in the Data folder.

# In[ ]:


import os
pospath = os.path.join('..','..','data','positive.txt')
negpath = os.path.join('..','..','data','negative.txt')

with open(negpath, "r") as f:
    negText = f.read()
negTokens = negText.split("\n") # This splits the text file into tokens on the new line character
negTokens[-1:] = [] # This strips out the final empty item
print(negTokens[-10:])

with open(pospath, "r") as f:
    posText = f.read()
posTokens = posText.split("\n") # This splits the text file into tokens on the new line character
posTokens[-1:] = [] # This strips out the final empty item
print(posTokens[-10:])


# In[ ]:


tweetpath = os.path.join('..','..','data','obama_tweets.txt')

with open(tweetpath, "r") as f:
    tweetsText = f.read()
tweetsTokens = tweetsText.split("\n") # This splits the text file into tokens on the new line character
tweetsTokens[-1:] = [] # This strips out the final empty item
print(tweetsTokens[:2])


# ### Tokenizing the tweets
# 
# Now we will create two functions. The first for tokenizing a tweet, the second for calculating positive/negative words.

# In[ ]:


import re
def tokenizer(theText):
    theTokens = re.findall(r'\b\w[\w-]*\b', theText.lower())
    return theTokens

def calculator(theTweet):
    # Count positive words
    numPosWords = 0
    theTweetTokens = tokenizer(theTweet)
    for word in theTweetTokens:
        if word in posTokens:
            numPosWords += 1
            
    # Count negative words
    numNegWords = 0
    for word in theTweetTokens:
        if word in negTokens:
            numNegWords += 1
    
    sum = (numPosWords - numNegWords)
    return sum



# In[ ]:


# a test liine!
print(calculator('Obama has called wrong wrong the GOP budget social Darwinism. Nice try, but they believe in social creationism.'))


# ### Analyzing tweets
# 
# Now we will use the calculator to calculate how many positive and negative tweets.
# 
# *Note:* that you can set a **threshold** for the number of words for a Tweet to be considered positive or negative.

# In[ ]:


# Here we set up the thresholds
posi = 1 # This means there have to be more than 1 positive word
nega = 0 # This means there has to be more than 1 negative words

# Here we prime our variables
numTweets = 0
numPosTweets = 0
numNegTweets = 0
numNeutTweets = 0

# This loop goes through all the Tweets and calculates if sums the number of positive or negative ones.
for tweet in tweetsTokens:
    calc = calculator(tweet)
    if calc > posi:
        numPosTweets += 1
        numTweets += 1
    elif calc < nega:
        numNegTweets += 1
        numTweets += 1
    else:
        numNeutTweets += 1
        numTweets += 1

# This prints out the results        
print("Total: " + str(numTweets) + "\n" + "Positive: " + str(numPosTweets) + "\n" + "Neutral: " + str(numNeutTweets) + "\n" + "Negative: " +str(numNegTweets))


# ### Gathering positive tweets
# 
# This will gather all examples of positive tweets.

# In[ ]:


# Here we set up the threshold.
posi = 1 # This means there have to be more than 1 positive word
numberWanted = 4 # Here you decide how many tweets you want

# Here we prime our variables
numTweets = 0
numPosTweets = 0
posiTweetList = []

# This loop goes through all the Tweets and calculates if sums the number of positive or negative ones.
for tweet in tweetsTokens:
    calc = calculator(tweet)
    if calc > posi and numPosTweets < numberWanted:
        numPosTweets += 1
        posiTweetList.append(tweet)

print(posiTweetList)


# ### Gathering negative tweets
# 
# This will gather examples of negative tweets.

# In[ ]:


# Here we set up the threshold.
nega = -1 # This means there have to be more than 1 positive word
numberWanted = 4 # Here you decide how many tweets you want

# Here we prime our variables
numTweets = 0
numNegTweets = 0
negaTweetList = []

# This loop goes through all the Tweets and calculates if sums the number of positive or negative ones.
for tweet in tweetsTokens:
    calc = calculator(tweet)
    if calc < nega and numNegTweets < numberWanted:
        numNegTweets += 1
        negaTweetList.append(tweet)

print(negaTweetList)


# ### Testing a tweet
# 
# Here you can take a tweet and test it to see how many positive or negative words it has.

# In[ ]:


tweetToCalc = input("What is the tweet to calculate? ")
print(calculator(tweetToCalc))


# ### Gathering and plotting positive and negative words
# 
# This will gather the words that are positive in the tweets and tabulate them.

# In[ ]:


import re
posWordsList = []
negWordsList = []

def tokenizer(theText):
    theTokens = re.findall(r'\b\w[\w-]*\b', theText.lower())
    return theTokens

def wordsCalculator(theTweet):
    # Count positive words
    numPosWords = 0
    theTweetTokens = tokenizer(theTweet)
    for word in theTweetTokens:
        if word in posTokens:
            numPosWords += 1
            posWordsList.append(word)
            
    # Count negative words
    numNegWords = 0
    for word in theTweetTokens:
        if word in negTokens:
            numNegWords += 1
            negWordsList.append(word)

tweet2Process = input("What tweet do you want to process? ")
wordsCalculator(tweet2Process)
print("Positive words: " + str(posWordsList[:10]))
print("Negative words: " + str(negWordsList[:10]))


# ### Gathering and plotting all positive and negative words

# In[ ]:


import re

# Here we set up the thresholds
posi = 1 # This means there must be more than 1 positive word
nega = 0 # This means there must be more than 1 negative words

# Here we prime our variables
posWordsList = []
negWordsList = []
numTweets = 0
numPosTweets = 0
numNegTweets = 0
numNeutTweets = 0

def wordsGathering(theTweet):
    # Count positive words
    numPosWords = 0
    theTweetTokens = tokenizer(theTweet)
    for word in theTweetTokens:
        if word in posTokens:
            numPosWords += 1
            posWordsList.append(word)
            
    # Count negative words
    numNegWords = 0
    for word in theTweetTokens:
        if word in negTokens:
            numNegWords += 1
            negWordsList.append(word)       
    
    sum = (numPosWords - numNegWords)
    return sum

# This loop goes through all the Tweets and calculates if sums the number of positive or negative ones.
for tweet in tweetsTokens:
    calc = wordsGathering(tweet)
    if calc > posi:
        numPosTweets += 1
        numTweets += 1
    elif calc < nega:
        numNegTweets += 1
        numTweets += 1
    else:
        numNeutTweets += 1
        numTweets += 1

print("Positive words: " + str(len(posWordsList)))
print("Negative words: " + str(len(negWordsList)))


# In[ ]:


# First let's plot the positive words.

import nltk, matplotlib

posDist = nltk.FreqDist(posWordsList)
posDist.tabulate(10)

get_ipython().magic('matplotlib inline')
posDist.plot(25, title="Top Positive Words")


# In[ ]:


# Now the negative words.

import nltk, matplotlib

negDist = nltk.FreqDist(negWordsList)
negDist.tabulate(10)

get_ipython().magic('matplotlib inline')
negDist.plot(25, title="Top Negative Words")

