
# coding: utf-8

# # <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Working with Data: More EDA!</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Longcan Huang</span>, GRA, CRMDA, <a class="author_email" href="mailto:#"> &lt;longcan@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> Pandas, EDA, tables, data
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # EDA of Tabular Data
# 
# Now that we've covered the basics of `pandas`, let's try to apply it to a dataset and see what we can see. We're back to Exploratory Data Analysis, or EDA.
# 
# Like our first EDA notebook, this one is adapted from Ted Underwood. You'll notice that we **reintroduce** some `pandas` concepts, becuase they went past like a whirlwind in the previous notebook!

# # Case Study - Movie Ratings on Fandango website

# ## Background
# 
# Our case study is based on the story *[Be Suspicious Of Online Movie Ratings, Especially Fandango’s](https://fivethirtyeight.com/features/fandango-movies-ratings/)* published on **[FiveThirtyEight](https://fivethirtyeight.com)**. A colleague of the author ([Walt Hickey](https://fivethirtyeight.com/contributors/walt-hickey/)) became suspicious of the [Fandango's](https://www.fandango.com) movie ratings because he noticed a bad movie is rated very high on the website.

# ## Research Questions
# 
# - Are the movie ratings on Fandango's website problematic?
#     - Does the Fandango's movie ratings follow the normal distribution?
#     - When compared to the other movie rating sources, are Fandango's movie ratings inflated?

# # Preparation

# ## Import Modules

# In[ ]:


import pandas as pd
import numpy as np


# ## Load the Data
# 
# - The dataset was created for **[FiveThirtyEight](https://fivethirtyeight.com)**. We'll use the `os` module to create a relative path, as we did with text files. In this case, the file name ends with '.csv,' because it's a "comma-separated" file.

# In[ ]:


import os
relativepath = os.path.join('..', '..', 'data', 'fandango.csv')
print(relativepath)


# - Now let's read in the whole table. The object `movies` is a `pandas` `DataFrame`, as we can see by passing the object to the `type()` function.

# In[ ]:


movies = pd.read_csv(relativepath)
type(movies)


# #  First Examination

# - A `pandas` `DataFrame` has an attribute `shape`, which returns a `tuple` of the number of rows (observations) and the number of columns (variables).

# In[ ]:


movies.shape


# - *Note: An attribute is different from a method, in that it does not take additional arguments.*

# In[ ]:


movies.shape()


# - You can print out the first `n` rows and the last `n` rows by using the `head(n)` and the `tail(n)` method to get a sense of the data, though that isn't terribly pretty with a DataFrame like this one that has 23 columns! By default, `n = 5`.

# In[ ]:


movies.head(100) # get the first five rows


# In[ ]:


movies.tail(3) # get the last five rows


# In[ ]:


movies.describe()


# # What are the variables?

# - If we want to know the variables as indicated by the column names, we can use the `columns` attribute of a `DataFrame`.

# In[ ]:


movies.columns


# - As shown, the type of `columns` attribute is `Index`, a `pandas` object resembling a one-dimentional array.

# ## The `pandas Index` object

# - The `pandas Index` object has the attribute `values` of type `list` and the attribute `dtype`.

# In[ ]:


print(movies.columns.values,
      '\n\n',
      movies.columns.dtype)


# - The variables are ratings from IMDB, Rotten Tomatoes, Metacritic, and Fandango. For the detailed descriptions of each variable, check out this [link](https://github.com/fivethirtyeight/data/tree/master/fandango).

# In[ ]:


movies.dtypes


# ## Take a peek at a single variable
# - Introducing `pandas` `Series`

# ## Create a `Series` from a column

# - by **column name**
#         DataFrame['column_name']

# In[ ]:


fandango_ratings = movies['Fandango_Ratingvalue'] # gives us all the fandango scores for the movies.
fandango_ratings.head()


# - by **index**
#         DataFrame.iloc[:, col_index]

# In[ ]:


movies.iloc[:, 7].head()


# ## Check the dtype of a `Series`

# In[ ]:


fandango_ratings.dtypes


# In[ ]:


fandango_ratings.dtype


# ## Check the distribution of a `Series`

# In[ ]:


fandango_ratings.describe()


# ## Exercise 1
# - create a `Series` named `fandango_votes` from the column 'Fandango_votes'
# - examine the distribution of the newly created object `fandango_votes` by using the `.describe()` method

# ## I'm so nosy!
# 
# - What are the movies that have the most and the least reviews?

# ## I'm so nosy!
# 
# - What movies are rated the lowest and the highest on Rotten Tomatoes?

# ## Subset the movies with 35+ reviews, i.e. remove invalid observations

# In[ ]:


thirtyfiveplus = movies['Fandango_votes'] > 35
thirtyfiveplus.value_counts()


# ## Slicing a `DataFrame` by **rows**

# - `DataFrame[condition]`

# In[ ]:


movies_thirtyfiveplus = movies[thirtyfiveplus]
movies_thirtyfiveplus.shape


# - `DataFrame.loc[condition, :]`

# In[ ]:


movies.loc[thirtyfiveplus, :].shape


# ## Exercise 2
# - create a `Series` named `fandango_thirtyfiveplus_r` that contains all the fandango ratings that are based on more than 35 votes.

# ## I'm so nosy!
# 
# - What are the names of the 50 highest rated movies on IMDB?

# ## Is the distribution of the Fandango ratings skewed?

# In[ ]:


fandango_thirtyfiveplus_r.describe()


# ## Visualizing the distribution of values in a column
# 
# - The `plot` method of a `Series`
# 
# - The `plot` method of a `DataFrame`
# 
# - The `plot` method of a `plt` state
#     
# ##### [A full tutorial is linked here.](http://pandas.pydata.org/pandas-docs/version/0.24/user_guide/visualization.html)#####

# ## The `plot` method of a `Series`

# In[ ]:


get_ipython().run_line_magic('matplotlib', 'inline')
import matplotlib.pyplot as plt
fandango_thirtyfiveplus_r.plot(kind='hist',
                           # density=True,
                           # color = 'r',
                           # alpha = 0.5
                          )
# plt.xlim(0, 5) # sets the range for x-axis
# plt.xlabel("Fandango Stars")


# In[ ]:


fandango_thirtyfiveplus_r.plot(kind='box')
plt.ylim(0, 5) # sets the range for y-axis


# ## Are Fandango stars inflated by rounding?
# 
# - relation between the variable `Fandango_Ratingvalue` and the variable `Fandango_Stars`

# ## Visualizing relationships between columns
# 
# - The `plot` method of a `DataFrame`

# - To handle the relations of two continuous variables, it is usually more intuitive to use a scatterplot.

# In[ ]:


movies.plot(kind='scatter', x='Fandango_Ratingvalue', y='Fandango_Stars')


# In[ ]:


movies.boxplot(column='Fandango_Ratingvalue', by='Fandango_Stars')
plt.ylim(0, 5)
plt.title("")


# ## Introducing `scatter_matrix`

# In[ ]:


from pandas.plotting import scatter_matrix
scatter_matrix(movies[['RottenTomatoes', 'RottenTomatoes_User', 'Metacritic',
       'Metacritic_User', 'IMDB']], figsize=[10, 10])
plt.show()


# ## Calculate the *correlation coefficient* between variables

# The strength of a linear relationship between two variables can be represented as a *correlation coefficient.* We'll be calculating the [Pearson correlation coefficient](https://en.wikipedia.org/wiki/Pearson_correlation_coefficient) between pairs of columns. The code below gives you an example of how to calculate this. Note that you can select columns from a DataFrame using the syntax
# 
#     dataframe['columnname']
#     
# and test the linearity of their relationship.

# In[ ]:


from scipy.stats.stats import pearsonr
coef = pearsonr(movies['Metacritic'], movies['RottenTomatoes'])
coef


# The `pearsonr` function returns a 2-tuple where the first element is the correlation coefficient. The linear relationship between two variables when the coefficient is close to 1 (positively correlated) or -1 (negatively correlated). A coefficient close to zero indicates weak correlation.
# 
# The second element in the 2-tuple is the `p` value. The P-value is the probability that you would have found the current result if the correlation coefficient were in fact zero. Conventionally, if this probability is lower than 5%, the correlation coefficient is statistically significant.

# ## Exercise 3
# 
# - Are the ratings from `RottenTomatoes` and from `IMDB` correlated? If so, what is the coeffient and the *p*-value?

# ## Are Fandango ratings comparable with the movie ratings from other sites?

# In[ ]:


movies.columns


# ## Select the columns of interest

# In[ ]:


import re
cols = movies.columns.tolist()
comparisons = [i for i in cols if re.search('no(r|)m$', i)]
comparisons


# In[ ]:


comparisons_round = [i for i in cols if re.search('round$', i)]
comparisons_round


# ## Compare `Fandango_Stars` with the other sites

# In[ ]:


movies.boxplot(column = comparisons_round, by='Fandango_Stars', figsize=(10, 10))
plt.show()


# ## Compare `Fandango_Ratingvalue` with the other sites

# In[ ]:


colors = ['r', 'b', 'g', 'y', 'k']
for i, j in zip(comparisons, colors):
    plt.hist(movies[i], color=j, alpha=0.5, label=i, density=True)
plt.hist(movies['Fandango_Ratingvalue'], color='gray', alpha=0.5, density=True)
plt.legend()

