
# coding: utf-8

# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>
# 
# <table style="width:95%;">
#   <tr style="background-color: #fff; font-family:'Helvetica'; font-size:120%">
#     <td style="padding-right:2em;">
#         <img src="https://crmda.dept.ku.edu/images/logoleft.png" alt="left side logo" style="width:60; height:60;">
#     </td>
#     <td style="width:60%; text-align:left; ">
#         <h1 style="text-align:center;">Python Data Science</h1>
#         <h3 style="text-align:center;">Working with Data: Some More Data Analysis</h3>
#         <div style=" display: block; margin-top: 1em; margin-bottom: 1em; border-style: solid; border-width: 1px;"></div>
#         <span style="font-weight:bold;">Dr. Jonathan Lamb</span>, Faculty Fellow, CRMDA, <a class="author_email" href="mailto:#"> &lt;jonathanplamb@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Longcan Huang</span>, GRA, CRMDA, <a class="author_email" href="mailto:#"> &lt;longcan@ku.edu&gt;</a><br><br>
#         <span style="font-weight:bold;">Keywords:</span> pandas, data visualization, data science
#     </td>
#     <td style="padding-left:2em;">
#     <img src="https://crmda.dept.ku.edu/images/logo-vert.png" alt="right side logo" style="width:60; height:60;">
#     </td>
# </tr>
# </table>
# 
# <div style=" display: block; margin-top: 2em; margin-bottom: 2em; border-style: solid; border-color: #0051ba; border-width: 2px;"></div>

# # Data analysis with `pandas`
# 
# Let's keep this movie theme going, and explore a different data set related to the movie industry. 
# 
# The questions we'll pose are based on [a parallel analysis by Albert Y. Kim, Chester Ismay, and Jennifer Chunn.](https://mran.microsoft.com/web/packages/fivethirtyeight/vignettes/bechdel.html) This notebook is adapted from a lesson by Ted Underwood of the University of Illinois.

# # Does gender bias exist in the movie business?
# 
# [The Dollar-And-Cents Case Against Hollywood’s Exclusion of Women](https://fivethirtyeight.com/features/the-dollar-and-cents-case-against-hollywoods-exclusion-of-women/)

# ## Background
# The Bechdel test originated from a 1985 strip by [Alison Bechdel](http://dykestowatchoutfor.com/). It has since been used to assess the gender bias in the movie industry. The test consists of three components:
# 
# 1. There are at least two named women in the picture;
# 2. They have a conversation with each other at some point;
# 3. That conversation isn’t about a male character.
# 
# A movie passes the bechdel test when it satisfies the three criteria.

# ## Research Questions
# - Does the bias exist?
#     - What is the distribution of the bechdel test results?
#     - Does the movies that fail the bechdel test get more budget?
# - Is the bias getting better or worse?
#      - Does the distribution change throughout the years?
# - Is the bias justified?
#     - Do movies that pass make more money per dollar?

# # Import the modules

# In[ ]:


get_ipython().run_line_magic('matplotlib', 'inline')
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


# # Import the data
# The data are collected from [bechdeltest.com](http://bechdeltest.com/) and [the-numbers.com](http://www.the-numbers.com/). The site bechdeltest.com provides detailed analyses for 5,000 films as to whether they pass the bechdel test. The site the-numbers.com on the other hand provides the financial information for about 4,500 films, including their budget and box office.

# In[ ]:


relativepath = os.path.join('..', '..', 'data', 'movies.csv')
bechdel = pd.read_csv(relativepath)
bechdel.head()


# In[ ]:


bechdel.describe(include='all')


# # Filter the data
# The intersection of The-Numbers and BechdelTest was a set of 1,615 films released between 1990 and 2013.
# - Focus only on films from 1990 to 2013

# In[ ]:


bechdel90_13 = bechdel.loc[(bechdel['year'] > 1989) & (bechdel['year'] < 2014)]
bechdel90_13.sample(5)


# # What are my variables?

# In[ ]:


bechdel90_13.dtypes


# # How are the test results distributed?

# In[ ]:


bechdel90_13['binary'].value_counts()


# In[ ]:


bechdel90_13['binary'].value_counts().plot(
    kind='bar'
)


# # Recode the variable
# 
# - For each film, if the clean_test returns 'ok' or 'dubious', recode it as `PASS`. Otherwise, recode it as `FAIL`.
# - [`DataFrame.apply(function)`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.apply.html)

# In[ ]:


bechdel90_13['clean_test'].value_counts()


# In[ ]:


bechdel90_13['clean_test'].isin(['ok', 'dubious']);


# In[ ]:


generous = bechdel90_13['clean_test'].apply(lambda x: 'PASS' if x in ['ok', 'dubious'] else 'FAIL')
# generous = bechdel90_13.apply(lambda x: 'PASS' if x['clean_test'] in ['ok', 'dubious'] else 'FAIL', axis=1)
generous.head()


# # Create new columns
# - use the `assign()` method

# In[ ]:


bechdel90_13 = bechdel90_13.assign(generous=generous)
# bechdel['generous'] = bechdel90_13['clean_test'].apply(lambda x: 'PASS' if x in ['ok', 'dubious'] else 'FAIL')
bechdel90_13.head()


# In[ ]:


bechdel90_13.generous.value_counts()


# In[ ]:


bechdel90_13.generous.value_counts().plot(kind='bar')


# # Do the movies that fail the test get more budget?
# - Compare the budgets of the movies that pass the test and the budgets of the movies that fail the test
#     1. group the movies based on the test results
#     2. each group should have a column containing the budgets for all the movies in this group
#     3. use the median of all the budgets in each group for comparison
# - Use the variable `generous` for test results
# - Use the variable `budget_2013$` for the overall budget

# ## Introduce `.groupby(column)`
# 
# `Pandas` has a `groupby()` method for achieving the splitting. Using `groupby()` creates a new object, called a `GroupBy`.

# In[ ]:


bechdel90_13.groupby('generous')['budget_2013$'].aggregate(np.median)
# bechdel90_13.groupby('generous').aggregate(np.median)['budget_2013$']


# In[ ]:


bechdel90_13['budget_2013$'].median()


# ## Understanding `.groupby(column)`
# You can think of this process of dividing the dataframe into several sub-dataframes based on the values of the given column. For each sub-dataframe, you can perform an operation and the result will be a `Series`. Now you have condensed several sub-dataframes into several `Series`, which is basically a `DataFrame` when combined with the column values as the index.

# In[ ]:


by_year = bechdel90_13.groupby('year')
print(type(by_year))
by_year.count()


# When operating on a `GroupBy` object, `aggregate()` is a common function to use. It takes a function as its argument. Some of the functions that can be passed in include `np.sum, np.mean, np.median,` and `np.std` (standard deviation). The prefix `np` is there to acknowledge that many of these are actually Numpy functions.

# In[ ]:


mean_budget = by_year[['budget', 'budget_2013$']].aggregate(np.mean)
mean_budget.head()


# ## Exercise 1
# 
# - Calculate the median of the budget for each group classified by the results of `clean_test`.

# ## Change the ordering of the index

# In[ ]:


budget_median = budget_median.reindex(['ok', 'dubious', 'men', 'notalk', 'nowomen'])
budget_median


# In[ ]:


from IPython.display import Image
from IPython.core.display import HTML
Image(url= "https://fivethirtyeight.com/wp-content/uploads/2014/04/hickey-bechdel-2.png?w=575")


# In[ ]:


# rects = plt.barh(budget_median.index, budget_median)
# for rect in rects:
#     width = rect.get_width()
#     plt.annotate(
#         '${:,.2f}m'.format(width/1000000),
#         xy=(width, rect.get_y() + rect.get_height()/2),
#         family='monospace'
#     )
# plt.box(False)
# plt.ticklabel_format(useOffset=False, axis='x', style='plain')
# plt.tick_params(top=False, bottom=False, left=False, right=False, labelleft=True, labelbottom=False)
# plt.title(
#     'Median Budget For Films Since 1990',
#     fontsize=20,
#     fontweight='bold',
#     family='serif',
#     loc='left'
# )
# ax = plt.gca()
# budget_median.index.dict = {
#     'ok': 'Passed Bechdel Test', 
#     'dubious': 'Barely passed Bechedel Test',
#     'men': 'Women only talk about men', 
#     'notalk': 'Women don\'t talk to each other', 
#     'nowomen': 'Fewer than two women',
# }
# ax.set_yticklabels(
#     [budget_median.index.dict[i] for i in budget_median.index.tolist()],
#     fontsize=14,
# )
# ax.invert_yaxis()
# # plt.show()


# # Is the gender bias getting better or worse?
# - group the movies by year
# - for each year compare the number of movies that fail and pass

# In[ ]:


bechdel90_13.groupby('year')['clean_test'].value_counts()


# In[ ]:


bechdel90_13.groupby('clean_test')['year'].value_counts(sort=False)


# In[ ]:


Image(url= "https://fivethirtyeight.com/wp-content/uploads/2014/04/hickey-bechdel-11.png?w=575")


# In[ ]:


year_cross_clean = pd.crosstab(bechdel90_13.year, bechdel90_13.clean_test).apply(lambda x: x/x.sum(), axis=1)
year_cross_clean.index


# In[ ]:


ordered_index = ['ok', 'dubious', 'men', 'notalk', 'nowomen']
colors = ('r', 'y', 'b', 'g', 'm')
# colors = ('#008fd5', '#6bb2d5', '#ffc9bf', '#ff9380', '#ff2700')
# colors = ('#008fd5', '#008fd5', 'k', 'k', 'k')
fig, ax = plt.subplots()
pos = list(year_cross_clean.index)
bottom = np.zeros(len(pos))
# for cat, color in zip(ordered_index, colors):
#     plt.bar(pos,
#             year_cross_clean[cat],
#             width=1,
#             bottom=bottom,
#             color=color
#            )
#     bottom += year_cross_clean[cat]
# plt.annotate(
#     'FAIL',
#     (1997, 0.7),
#     color='white',
#     fontsize=50,
#     family='monospace'
# )
# plt.annotate(
#     'PASS',
#     (1997, 0.2),
#     color='white',
#     fontsize=50,
#     family='monospace'
# )


# # Over the years, how do the movie budget change?

# In[ ]:


bechdel90_13.groupby('year').aggregate(np.median).plot(y='budget_2013$')

# failed_budget = bechdel90_13[bechdel90_13.binary == 'FAIL'].groupby('year').aggregate(np.median)['budget_2013$']
# passed_budget = bechdel90_13[bechdel90_13.binary == 'PASS'].groupby('year').aggregate(np.median)['budget_2013$']
# failed_budget.plot(label='fail')
# passed_budget.plot(label='pass')

# plt.legend(loc='upper right')
# plt.text(2010, 35000000, 'all')
# plt.text(2010, 60000000, 'fail')
# plt.text(2010, 16000000, 'pass')


# # Do movies that pass the bechedel test make more money or less money per dollar spent? Globally? Overseas? Domestic?
# 
# - For this purpose, we need to calculate a few variables first.
# 
#     1. `int_only`: international gross only
#             total gross (`intgross_2013$`) - domestic gross (`domgross_2013$`)
#     2. `roi_tot`: total return on investiment
#             total gross (`intgross_2013$`) / budget (`budget_2013$`)
#     3. `roi_dom`: domestic return on investiment
#             domestic gross (`domgross_2013$`) / budget (`budget_2013$`)
#     4. `roi_int`: international/overseas return on investiment
#             internation gross only (`int_only`) / budget (`budget_2013$`)

# In[ ]:


bechdel90_13 = bechdel90_13.assign(
    int_only=bechdel90_13['intgross_2013$'] - bechdel90_13['domgross_2013$']
)
bechdel90_13.head()


# In[ ]:


bechdel90_13 = bechdel90_13.assign(
    roi_tot=bechdel90_13['intgross_2013$'] / bechdel90_13['budget_2013$'],
    roi_dom=bechdel90_13['domgross_2013$'] / bechdel90_13['budget_2013$']
)
bechdel.head()


# ## Exercise 2
# - calculate the return on investment (ROI) of only the international gross and assign it to a new column named `roi_int`.

# In[ ]:


from IPython.display import Image
from IPython.core.display import HTML 
Image(url= "https://fivethirtyeight.com/wp-content/uploads/2014/04/hickey-bechdel-3.png?w=575")


# In[ ]:


median_by_test = bechdel90_13.groupby('clean_test').aggregate(np.median)
median_by_test = median_by_test.reindex(['ok', 'dubious', 'men', 'notalk', 'nowomen'])
median_roi = [median_by_test['roi_dom'], median_by_test['roi_int']]


# In[ ]:


fig, axes = plt.subplots(1, 2, sharey=True)
# def ax_barh(df, ax, i):
#     showleft = [True, False][i]
#     color = ['#4169e1', '#daa520'][i]
#     title = ['US and Canada', 'International Only'][i]
#     rects = ax.barh(df.index, df, color=color)
#     for rect in rects:
#         width = rect.get_width()
#         ax.annotate(
#             '${:,.2f}'.format(width),
#             xy=(width, rect.get_y() + rect.get_height()/2),
#             family='monospace'
#     )
#     ax.spines['top'].set_visible(False)
#     ax.spines['right'].set_visible(False)
#     ax.spines['bottom'].set_visible(False)
#     ax.spines['left'].set_visible(False)
#     ax.ticklabel_format(useOffset=False, axis='x', style='plain')
#     ax.tick_params(top=False, bottom=False, left=False, right=False, labelleft=showleft, labelbottom=False)
#     ax.set_title(
#         title,
#         fontsize=15,
#         fontweight='bold'
#     )

# for i, j in enumerate(median_roi):
#     ax_barh(j, axes[i], i)
    
# budget_median.index.dict = {
#     'ok': 'Passed Bechdel Test', 
#     'dubious': 'Barely passed Bechedel Test',
#     'men': 'Women only talk about men', 
#     'notalk': 'Women don\'t talk to each other', 
#     'nowomen': 'Fewer than two women',
# }
# ax1 = axes[0]
# ax1.set_yticklabels(
#     [budget_median.index.dict[i] for i in budget_median.index.tolist()],
#     fontsize=14,
#     )
# ax1.invert_yaxis()
# fig.suptitle(
#     'Dollors Earned for Every Dollar Spent',
#     fontsize=20,
#     fontweight='bold',
#     family='serif',
#     y=1.1,
#     ha='right'
# )

